#	include	"Standard.h"

int		StallPenaltyForBranchMissPredict	=	2;
int		StallPenaltyForValueMissPredict		=	2;
extern	double	ValuePredictionConfidenceThreshold;

bool
ExecutionUnit	::	Initialize (
	int		aIntegerUnits,
	int		aFPUnits,
	int		aLoadUnits,
	int		aStoreUnits,
	int		aBranchUnits,

	int		aIntegerQueueSize,
	int		aFPQueueSize,
	int		aLoadQueueSize,
	int		aStoreQueueSize,
	int		aBranchQueueSize,

	int		aIntegerPipelineDelay,
	int		aFPPipelineDelay,
	int		aLoadPipelineDelay,
	int		aStorePipelineDelay,
	int		aBranchPipelineDelay,

	CompletionUnit	*aCompletionUnit,
	int		aL1MissPenalty,
	int		aL2MissPenalty	)
	{
	Stalled = 0;
	L1MissPenalty = aL1MissPenalty;
	L2MissPenalty = aL2MissPenalty;

	IntegerUnits = aIntegerUnits;
	FPUnits = aFPUnits;
	LoadUnits = aLoadUnits;
	StoreUnits = aStoreUnits;
	BranchUnits = aBranchUnits;

	IntegerQueueSize = aIntegerQueueSize;	
	FPQueueSize = aFPQueueSize;	
	LoadQueueSize = aLoadQueueSize;	
	StoreQueueSize = aStoreQueueSize;	
	BranchQueueSize = aBranchQueueSize;	

	IntegerPipelineDelay = aIntegerPipelineDelay;
	FPPipelineDelay = aFPPipelineDelay;
	LoadPipelineDelay = aLoadPipelineDelay;
	StorePipelineDelay = aStorePipelineDelay;
	BranchPipelineDelay = aBranchPipelineDelay;

	IntegerQueueValid = 0;
	FPQueueValid = 0;
	LoadQueueValid = 0;
	StoreQueueValid = 0;
	BranchQueueValid = 0;

	IntegerInstructions = new Instruction[IntegerQueueSize];
	IntegerAddresses = new uint[IntegerQueueSize];
	memset(IntegerInstructions, 0, sizeof(Instruction)*IntegerQueueSize);
	memset(IntegerAddresses, 0, sizeof(uint)*IntegerQueueSize);

	FPInstructions = new Instruction[FPQueueSize];
	FPAddresses = new uint[FPQueueSize];
	memset(FPInstructions, 0, sizeof(Instruction)*FPQueueSize);
	memset(FPAddresses, 0, sizeof(uint)*FPQueueSize);

	LoadInstructions = new Instruction[LoadQueueSize];
	LoadAddresses = new uint[LoadQueueSize];
	memset(LoadInstructions, 0, sizeof(Instruction)*LoadQueueSize);
	memset(LoadAddresses, 0, sizeof(uint)*LoadQueueSize);

	StoreInstructions = new Instruction[StoreQueueSize];
	StoreAddresses = new uint[StoreQueueSize];
	memset(StoreInstructions, 0, sizeof(Instruction)*StoreQueueSize);
	memset(StoreAddresses, 0, sizeof(uint)*StoreQueueSize);

	BranchInstructions = new Instruction[BranchQueueSize];
	BranchAddresses = new uint[BranchQueueSize];
	memset(BranchInstructions, 0, sizeof(Instruction)*BranchQueueSize);
	memset(BranchAddresses, 0, sizeof(uint)*BranchQueueSize);

	int	Index;

#	define	AllocatePipeline(name, units, delay)	\
	name = new PipelineStage*[units];\
	for(Index = 0;Index<units;Index++)\
		{\
		name[Index] = new PipelineStage[delay];\
		memset(name[Index], 0, sizeof(PipelineStage) * delay);\
		}

	AllocatePipeline(PipelineOfInteger, IntegerUnits, IntegerPipelineDelay);
	AllocatePipeline(PipelineOfFP, FPUnits, FPPipelineDelay);
	AllocatePipeline(PipelineOfLoad, LoadUnits, LoadPipelineDelay);
	AllocatePipeline(PipelineOfStore, StoreUnits, StorePipelineDelay);
	AllocatePipeline(PipelineOfBranch, BranchUnits, BranchPipelineDelay);


	theCompletionUnit = aCompletionUnit;

	return ( true );
	}

int
ExecutionUnit	::	StallsDueToCache (
	NormalDistribution	l1,
	NormalDistribution	l2 )
	{
	// generate a uniform random number and a normally distributed
	// number.  if uniform <= normally distributed then a hit.
	// otherwise a miss!
	if (	RandomVariable :: GetUniformRandomVariable(1.0) <=
			RandomVariable :: GetNormalRandomVariable(l1)	)
		return (1);

	// Miss in L1!
	if (	RandomVariable :: GetUniformRandomVariable(1.0) <=
			RandomVariable :: GetNormalRandomVariable(l2)	)
		return ( L1MissPenalty + 1 );

	return ( L2MissPenalty + L1MissPenalty + 1);
	}

bool
ExecutionUnit	::	FreeSlots (
	int		*IntegerSlots,
	int		*FPSlots,
	int		*LoadSlots,
	int		*StoreSlots,
	int		*BranchSlots )
	{
	(*IntegerSlots) = (IntegerQueueSize - IntegerQueueValid);
	(*FPSlots) = (FPQueueSize - FPQueueValid);
	(*LoadSlots) = (LoadQueueSize - LoadQueueValid);
	(*StoreSlots) = (StoreQueueSize - StoreQueueValid);
	(*BranchSlots) = (BranchQueueSize - BranchQueueValid);
	return ( true );
	}

bool
ExecutionUnit	::	AddInstructions (
	Instruction		*aIntegerInstructions,
	uint			*aIntegerAddresses,
	int				NumberOfIntegerInstructions,
	Instruction		*aFPInstructions,
	uint			*aFPAddresses,
	int				NumberOfFPInstructions,
	Instruction		*aLoadInstructions,
	uint			*aLoadAddresses,
	int				NumberOfLoadInstructions,
	Instruction		*aStoreInstructions,
	uint			*aStoreAddresses,
	int				NumberOfStoreInstructions,
	Instruction		*aBranchInstructions,
	uint			*aBranchAddresses,
	int				NumberOfBranchInstructions )
	{
	memcpy ( & IntegerInstructions [ IntegerQueueValid ], 
		aIntegerInstructions,
		sizeof ( Instruction ) * NumberOfIntegerInstructions );
	memcpy ( & IntegerAddresses [ IntegerQueueValid ],
		aIntegerAddresses,
		sizeof ( uint ) * NumberOfIntegerInstructions );
	IntegerQueueValid += NumberOfIntegerInstructions;

	memcpy ( & FPInstructions [ FPQueueValid ], aFPInstructions,
		sizeof ( Instruction ) * NumberOfFPInstructions );
	memcpy ( & FPAddresses [ FPQueueValid ], aFPAddresses,
		sizeof ( uint ) * NumberOfFPInstructions );
	FPQueueValid += NumberOfFPInstructions;

	memcpy ( & LoadInstructions [ LoadQueueValid ],
		aLoadInstructions,
		sizeof ( Instruction ) * NumberOfLoadInstructions );
	memcpy ( & LoadAddresses [ LoadQueueValid ], aLoadAddresses,
		sizeof ( uint ) * NumberOfLoadInstructions );
	LoadQueueValid += NumberOfLoadInstructions;

	memcpy ( & StoreInstructions [ StoreQueueValid ],
		aStoreInstructions,
		sizeof ( Instruction ) * NumberOfStoreInstructions );
	memcpy ( & StoreAddresses [ StoreQueueValid ],
		aStoreAddresses,
		sizeof ( uint ) * NumberOfStoreInstructions );
	StoreQueueValid += NumberOfStoreInstructions;

	memcpy ( & BranchInstructions [ BranchQueueValid ],
		aBranchInstructions,
		sizeof ( Instruction ) * NumberOfBranchInstructions );
	memcpy ( & BranchAddresses [ BranchQueueValid ],
		aBranchAddresses,
		sizeof ( uint ) * NumberOfBranchInstructions );
	BranchQueueValid += NumberOfBranchInstructions;
	return ( true );
	}


bool
ExecutionUnit	::	SimulateOneClock (
	bool		*ResetBranch,
	uint		*ToAddress,
	int64		*ToCount,
	bool		*ResetDueToBranch )
	{
	// For each unit, execute it
	bool	Issued = false;
	double	Confidence, iConfidence;
	int64	Temp64;

	(*ResetBranch) = false;
	(*ResetDueToBranch) = false;

	Temp64 = 1;
	Temp64 = Temp64<<62;
	(*ToCount) = Temp64;
	
	if(Stalled)
		{
		--Stalled;
		return(true);
		}	
	int	FU, PS;
	int	Index, Index2;
	bool	ReadyToIssue;
	
	// Integer unit
	// first step: drain the pipeline

	///////////////////////////////////////////////////////////////////////////////
	//                           Integer
	///////////////////////////////////////////////////////////////////////////////

	for(FU=0;FU < IntegerUnits;FU++)
		{
		PS = 0;
		if ( PipelineOfInteger [ FU ] [ PS ].Valid )
			{
			if ( PipelineOfInteger[FU][PS].theInstruction.MissPredictedOutcome )
				{
				(*ResetBranch) = true;
				if( PipelineOfInteger[FU][PS].theInstruction.InstructionCount < (*ToCount) )
					{
					(*ToCount) = PipelineOfInteger[ FU ] [ PS ].theInstruction.InstructionCount;
					(*ToAddress) = PipelineOfInteger[ FU ] [ PS ].theAddress;
					}
				//printf("MISS PREDICT I!\n");
				}
			if(theCompletionUnit->CompleteInstruction (
				PipelineOfInteger [ FU ] [ PS ].theInstruction.InstructionCount,
				PipelineOfInteger [ FU ] [ PS ].Confidence,
				PipelineOfInteger [ FU ] [ PS ].Dependence0,
				PipelineOfInteger [ FU ] [ PS ].Dependence1 )	)
				{

#	define	ShiftPipeline(name, delay)	\
	memcpy ( & name[ FU ] [ 0 ],\
			 & name[ FU ] [ 1 ],\
			sizeof(PipelineStage) * (delay-1)	);\
	name[FU][delay-1].Valid = false;

				ShiftPipeline(PipelineOfInteger, IntegerPipelineDelay);
				}
			//else
			}
		else
			{
			ShiftPipeline(PipelineOfInteger, IntegerPipelineDelay);
			}
		}

	// next step, inject new instructions ( if possible )

	for(FU=0;FU < IntegerUnits;FU++)
		{
		PS = IntegerPipelineDelay - 1;
		if ( ! PipelineOfInteger[ FU ] [ PS ].Valid )
			{
			// search for an instruction to execute
			Issued = false;
			Index=0;
			while(Index<IntegerQueueValid)
				{
				ReadyToIssue = true;
				for(Index2=0;Index2<MaxInputDependencies;Index2++)
					{
					if ( IntegerInstructions[Index].InputDependencies[Index2] )
					if ( !theCompletionUnit->IsCompleted (
						IntegerInstructions[Index].InstructionCount -
						IntegerInstructions[Index].InputDependencies[Index2] ) )
						ReadyToIssue = false;
					}

				if ( ReadyToIssue )
					{
					// test to see if this is memoizable, if so, don't issue it, just complete it
					if ( (FU == 0) && (IntegerInstructions[Index].Memoizable) )
						{
						if(theCompletionUnit->CompleteInstruction (
							IntegerInstructions[Index].InstructionCount))
							{
							memcpy ( & IntegerInstructions [ Index ],
								 & IntegerInstructions [ Index + 1],
								sizeof ( Instruction ) * (IntegerQueueValid - Index - 1) );
							memcpy ( & IntegerAddresses [ Index ],
								 & IntegerAddresses [ Index + 1],
								sizeof ( uint ) * (IntegerQueueValid - Index - 1) );
							--IntegerQueueValid;
							}
						++Index;
						continue;
						}
					else
						{
#	define	InjectPipeline(name, inst, addr, conf, qv)	\
						memcpy ( & name [ FU ] [ PS ].theInstruction, & inst[ Index ], sizeof ( Instruction ) );\
						memcpy ( & name [ FU ] [ PS ].theAddress, & addr[ Index ], sizeof ( uint ) );\
						memcpy ( & inst [ Index ], & inst [ Index + 1], sizeof ( Instruction ) * (qv - Index - 1) );\
						memcpy ( & addr [ Index ], & addr [ Index + 1], sizeof ( uint ) * (qv - Index - 1) );\
						name [ FU ] [ PS ].Valid = true;\
						name [ FU ] [ PS ].Confidence = conf;\
						if ( name[FU][PS].theInstruction.InputDependencies[0] )\
							name [ FU ] [ PS ].Dependence0 = name[FU][PS].theInstruction.InstructionCount -\
								name[FU][PS].theInstruction.InputDependencies[0];\
						else\
							name [ FU ] [ PS ].Dependence0 = -1;\
						if ( name[FU][PS].theInstruction.InputDependencies[1] )\
							name [ FU ] [ PS ].Dependence1 = name[FU][PS].theInstruction.InstructionCount -\
								name[FU][PS].theInstruction.InputDependencies[1];\
						else\
							name [ FU ] [ PS ].Dependence1 = -1;\
						--qv;Issued=true;

						InjectPipeline(PipelineOfInteger, IntegerInstructions, IntegerAddresses, 1.0, IntegerQueueValid);
						}
					break;
					}
				++Index;
				}
			if(!Issued)
				{
				// o.k., go through the queue again, but this time searching for someone to value
				// speculate about... (that is value speculate about the inputs)!

				Index = 0;
				while(Index<IntegerQueueValid)
					{
					ReadyToIssue = true;
					Confidence = 1.0;
					for(Index2=0;Index2<MaxInputDependencies;Index2++)
						{
						if( IntegerInstructions[Index].InputDependencies[Index2] )
							{
							if ( !theCompletionUnit->IsCompletedWithConfidence (
								IntegerInstructions[Index].InstructionCount -
								IntegerInstructions[Index].InputDependencies[Index2],
								ValuePredictionConfidenceThreshold,
								&iConfidence) )
								ReadyToIssue = false;
							else
								Confidence = Confidence * iConfidence;
							}
						}
					if ( ReadyToIssue )
					if ( Confidence >= ValuePredictionConfidenceThreshold )
						{
						InjectPipeline(PipelineOfInteger,
							IntegerInstructions, IntegerAddresses, Confidence, IntegerQueueValid);
						break;
						}
					++Index;
					}
				}
			}
		}

	///////////////////////////////////////////////////////////////////////////////
	//                           FP
	///////////////////////////////////////////////////////////////////////////////

	for(FU=0;FU < FPUnits;FU++)
		{
		PS = 0;
		if ( PipelineOfFP[ FU ] [ PS ].Valid )
			{
			if ( PipelineOfFP[FU][PS].theInstruction.MissPredictedOutcome )
				{
				(*ResetBranch) = true;
				if( PipelineOfFP[FU][PS].theInstruction.InstructionCount < (*ToCount) )
					{
					(*ToCount) = PipelineOfFP[ FU ] [ PS ].theInstruction.InstructionCount;
					(*ToAddress) = PipelineOfFP[ FU ] [ PS ].theAddress;
					}
				//printf("MISS PREDICT F!\n");
				}
			if(theCompletionUnit->CompleteInstruction (
				PipelineOfFP [ FU ] [ PS ].theInstruction.InstructionCount,
				PipelineOfFP [ FU ] [ PS ].Confidence,
				PipelineOfFP [ FU ] [ PS ].Dependence0,
				PipelineOfFP [ FU ] [ PS ].Dependence1 )	)
				{
				ShiftPipeline(PipelineOfFP, FPPipelineDelay);
				}
			}
		else
			{
			ShiftPipeline(PipelineOfFP, FPPipelineDelay);
			}
		}

	// next step, inject new instructions ( if possible )

	for(FU=0;FU < FPUnits;FU++)
		{
		PS = FPPipelineDelay - 1;
		if ( ! PipelineOfFP[ FU ] [ PS ].Valid )
			{
			// search for an instruction to execute
			Issued = false;
			Index=0;
			while(Index<FPQueueValid)
				{
				ReadyToIssue = true;
				for(Index2=0;Index2<MaxInputDependencies;Index2++)
					{
					if ( FPInstructions[Index].InputDependencies[Index2] )
					if ( !theCompletionUnit->IsCompleted (
						FPInstructions[Index].InstructionCount -
						FPInstructions[Index].InputDependencies[Index2] ) )
						ReadyToIssue = false;
					}

				if ( ReadyToIssue )
					{
					
					if ( (FU==0) && (FPInstructions[Index].Memoizable) )
						{
						if(theCompletionUnit->CompleteInstruction (
							FPInstructions[Index].InstructionCount))
							{
							memcpy ( & FPInstructions [ Index ],
								 & FPInstructions [ Index + 1],
								sizeof ( Instruction ) * (FPQueueValid - Index - 1) );
							memcpy ( & FPAddresses [ Index ],
								 & FPAddresses [ Index + 1],
								sizeof ( uint ) * (FPQueueValid - Index - 1) );
							--FPQueueValid;
							}
						++Index;
						continue;
						}
					else
						{
						InjectPipeline(PipelineOfFP, FPInstructions, FPAddresses, 1.0, FPQueueValid);
						}
					break;
					}

				// add some value speculation code?

				++Index;
				}
			if(!Issued)
				{
				// o.k., go through the queue again, but this time searching for someone to value
				// speculate about... (that is value speculate about the inputs)!

				Index = 0;
				while(Index<FPQueueValid)
					{
					ReadyToIssue = true;
					Confidence = 1.0;
					for(Index2=0;Index2<MaxInputDependencies;Index2++)
						{
						if( FPInstructions[Index].InputDependencies[Index2] )
							{
							if ( !theCompletionUnit->IsCompletedWithConfidence (
								FPInstructions[Index].InstructionCount -
								FPInstructions[Index].InputDependencies[Index2],
								ValuePredictionConfidenceThreshold,
								&iConfidence) )
								ReadyToIssue = false;
							else
								Confidence = Confidence * iConfidence;
							}
						}
					if ( ReadyToIssue )
					if ( Confidence >= ValuePredictionConfidenceThreshold )
						{
						InjectPipeline(PipelineOfFP,
							FPInstructions, FPAddresses, Confidence, FPQueueValid);
						break;
						}
					++Index;
					}
				}
			}
		}

	///////////////////////////////////////////////////////////////////////////////
	//                           Load
	///////////////////////////////////////////////////////////////////////////////

	for(FU=0;FU < LoadUnits;FU++)
		{
		PS = 0;
		if(PipelineOfLoad[FU][PS].theInstruction.ExecutionState)
			--PipelineOfLoad[FU][PS].theInstruction.ExecutionState;
		if(!PipelineOfLoad[FU][PS].theInstruction.ExecutionState)
			{
			if ( PipelineOfLoad[ FU ] [ PS ].Valid )
				{
				if ( PipelineOfLoad[FU][PS].theInstruction.MissPredictedOutcome )
					{
					(*ResetBranch) = true;
					if( PipelineOfLoad[FU][PS].theInstruction.InstructionCount < (*ToCount) )
						{
						(*ToCount) = PipelineOfLoad[ FU ] [ PS ].theInstruction.InstructionCount;
						(*ToAddress) = PipelineOfLoad[ FU ] [ PS ].theAddress;
						}
					//printf("MISS PREDICT L!\n");
					}
				if(theCompletionUnit->CompleteInstruction (
					PipelineOfLoad [ FU ] [ PS ].theInstruction.InstructionCount ))
					{
					ShiftPipeline(PipelineOfLoad, LoadPipelineDelay);
					}
				}
			else
				{
				ShiftPipeline(PipelineOfLoad, LoadPipelineDelay);
				}
			}
		}

	// next step, inject new instructions ( if possible )

	for(FU=0;FU < LoadUnits;FU++)
		{
		PS = LoadPipelineDelay - 1;
		if ( ! PipelineOfLoad[ FU ] [ PS ].Valid )
			{
			Issued = false;
			// search for an instruction to execute
			Index=0;
			while(Index<LoadQueueValid)
				{
				ReadyToIssue = true;
				for(Index2=0;Index2<MaxInputDependencies;Index2++)
					{
					if ( LoadInstructions[Index].InputDependencies[Index2] )
					if ( !theCompletionUnit->IsCompleted (
						LoadInstructions[Index].InstructionCount -
						LoadInstructions[Index].InputDependencies[Index2] ) )
						ReadyToIssue = false;
					}

				if ( ReadyToIssue )
					{
					InjectPipeline(PipelineOfLoad, LoadInstructions, LoadAddresses, 1.0, LoadQueueValid);
					if ( PipelineOfLoad[ FU ] [ PS ].theInstruction.FunctionalUnitRequirements == Load)
						PipelineOfLoad[ FU ] [ PS ].theInstruction.ExecutionState =
							StallsDueToCache ( 
								PipelineOfLoad[FU][PS].theInstruction.L1DataHitPredictability,
								PipelineOfLoad[FU][PS].theInstruction.L2DataHitPredictability );
					else
						PipelineOfLoad[ FU ] [ PS ].theInstruction.ExecutionState = 0;
					break;
					}

				// add some value speculation code?

				++Index;
				}
			if(!Issued)
				{
				// o.k., go through the queue again, but this time searching for someone to value
				// speculate about... (that is value speculate about the inputs)!

				Index = 0;
				while(Index<LoadQueueValid)
					{
					ReadyToIssue = true;
					Confidence = 1.0;
					for(Index2=0;Index2<MaxInputDependencies;Index2++)
						{
						if( LoadInstructions[Index].InputDependencies[Index2] )
							{
							if ( !theCompletionUnit->IsCompletedWithConfidence (
								LoadInstructions[Index].InstructionCount -
								LoadInstructions[Index].InputDependencies[Index2],
								ValuePredictionConfidenceThreshold,
								&iConfidence) )
								ReadyToIssue = false;
							else
								Confidence = Confidence * iConfidence;
							}
						}
					if ( ReadyToIssue )
					if ( Confidence >= ValuePredictionConfidenceThreshold )
						{
						InjectPipeline(PipelineOfLoad,
							LoadInstructions, LoadAddresses, Confidence, LoadQueueValid);
						if ( PipelineOfLoad[ FU ] [ PS ].theInstruction.FunctionalUnitRequirements == Load)
							PipelineOfLoad[ FU ] [ PS ].theInstruction.ExecutionState =
								StallsDueToCache ( 
									PipelineOfLoad[FU][PS].theInstruction.L1DataHitPredictability,
									PipelineOfLoad[FU][PS].theInstruction.L2DataHitPredictability );
						else
							PipelineOfLoad[ FU ] [ PS ].theInstruction.ExecutionState = 0;
						break;
						}
					++Index;
					}
				}
			}
		}

/**************************************************************************************
	///////////////////////////////////////////////////////////////////////////////
	//                           Store
	///////////////////////////////////////////////////////////////////////////////

	for(FU=0;FU < StoreUnits;FU++)
		{
		PS = 0;
		if(PipelineOfStore[FU][PS].theInstruction.ExecutionState)
			--PipelineOfStore[FU][PS].theInstruction.ExecutionState;
		if(!PipelineOfStore[FU][PS].theInstruction.ExecutionState)
			{
			if ( PipelineOfStore[ FU ] [ PS ].Valid )
				{
				if ( PipelineOfStore[FU][PS].theInstruction.MissPredictedOutcome )
					{
					(*ResetBranch) = true;
					if( PipelineOfStore[FU][PS].theInstruction.InstructionCount < (*ToCount) )
						{
						(*ToCount) = PipelineOfStore[ FU ] [ PS ].theInstruction.InstructionCount;
						(*ToAddress) = PipelineOfStore[ FU ] [ PS ].theAddress;
						}
					//printf("MISS PREDICT S!\n");
					}
				if(theCompletionUnit->CompleteInstruction (
					PipelineOfStore[ FU ] [ PS ].theInstruction.InstructionCount ))
					{
					ShiftPipeline(PipelineOfStore, StorePipelineDelay);
					}
				}
			else
				{
				ShiftPipeline(PipelineOfStore, StorePipelineDelay);
				}
			}
		}

	// next step, inject new instructions ( if possible )

	for(FU=0;FU < StoreUnits;FU++)
		{
		PS = StorePipelineDelay - 1;
		if ( ! PipelineOfStore[ FU ] [ PS ].Valid )
			{
			Issued = false;
			// search for an instruction to execute
			Index=0;
			while(Index<StoreQueueValid)
				{
				ReadyToIssue = true;
				for(Index2=0;Index2<MaxInputDependencies;Index2++)
					{
					if ( StoreInstructions[Index].InputDependencies[Index2] )
					if ( !theCompletionUnit->IsCompleted (
						StoreInstructions[Index].InstructionCount -
						StoreInstructions[Index].InputDependencies[Index2] ) )
						ReadyToIssue = false;
					}

				if ( ReadyToIssue )
					{
					InjectPipeline(PipelineOfStore, StoreInstructions, StoreAddresses, 1.0, StoreQueueValid);
					PipelineOfStore[ FU ] [ PS ].theInstruction.ExecutionState =
						StallsDueToCache (
							PipelineOfStore[FU][PS].theInstruction.L1DataHitPredictability,
							PipelineOfStore[FU][PS].theInstruction.L2DataHitPredictability );
					break;
					}

				// add some value speculation code?

				++Index;
				}
			if(!Issued)
				{
				// o.k., go through the queue again, but this time searching for someone to value
				// speculate about... (that is value speculate about the inputs)!

				Index = 0;
				while(Index<StoreQueueValid)
					{
					ReadyToIssue = true;
					Confidence = 1.0;
					for(Index2=0;Index2<MaxInputDependencies;Index2++)
						{
						if( StoreInstructions[Index].InputDependencies[Index2] )
							{
							if ( !theCompletionUnit->IsCompletedWithConfidence (
								StoreInstructions[Index].InstructionCount -
								StoreInstructions[Index].InputDependencies[Index2],
								ValuePredictionConfidenceThreshold,
								&iConfidence) )
								ReadyToIssue = false;
							else
								Confidence = Confidence * iConfidence;
							}
						}
					if ( ReadyToIssue )
					if ( Confidence >= ValuePredictionConfidenceThreshold )
						{
						InjectPipeline(PipelineOfStore,
							StoreInstructions, StoreAddresses, Confidence, StoreQueueValid);
						PipelineOfLoad[ FU ] [ PS ].theInstruction.ExecutionState =
							StallsDueToCache ( 
								PipelineOfStore[FU][PS].theInstruction.L1DataHitPredictability,
								PipelineOfStore[FU][PS].theInstruction.L2DataHitPredictability );
							break;
						}
					++Index;
					}
				}
			}
		}
*****************************************************************************/

	///////////////////////////////////////////////////////////////////////////////
	//                           Branch
	///////////////////////////////////////////////////////////////////////////////

	for(FU=0;FU < BranchUnits;FU++)
		{
		PS = 0;
		if ( PipelineOfBranch[ FU ] [ PS ].Valid )
			{
			if ( PipelineOfBranch[FU][PS].theInstruction.MissPredictedOutcome )
				{
				(*ResetBranch) = true;
				if( PipelineOfBranch[FU][PS].theInstruction.InstructionCount < (*ToCount) )
					{
					(*ToCount) = PipelineOfBranch[ FU ] [ PS ].theInstruction.InstructionCount;
					(*ToAddress) = PipelineOfBranch[ FU ] [ PS ].theAddress;
					}
				//printf("MISS PREDICT B!\n");
				}
			if(theCompletionUnit->CompleteInstruction (
				PipelineOfBranch[ FU ] [ PS ].theInstruction.InstructionCount ))
				{
				if (PipelineOfBranch[ FU ] [ PS ].theInstruction.IsAMissSpeculationDueToBranch)
					{
					(*ResetBranch) = true;
					if( (PipelineOfBranch[FU][PS].theInstruction.InstructionCount+1) < (*ToCount) )
						{
						(*ToCount) = PipelineOfBranch[ FU ] [ PS ].theInstruction.InstructionCount + 1;
						(*ToAddress) = PipelineOfBranch[ FU ] [ PS ].theInstruction.FixupTarget;
						(*ResetDueToBranch) = true;
						}
					}
				ShiftPipeline(PipelineOfBranch, BranchPipelineDelay);
				}
			}
		else
			{
			ShiftPipeline(PipelineOfBranch, BranchPipelineDelay);
			}
		}

	// next step, inject new instructions ( if possible )

	for(FU=0;FU < BranchUnits;FU++)
		{
		PS = BranchPipelineDelay - 1;
		if ( ! PipelineOfBranch[ FU ] [ PS ].Valid )
			{
			Issued = false;
			// search for an instruction to execute
			Index=0;
			while(Index<BranchQueueValid)
				{
				ReadyToIssue = true;
				for(Index2=0;Index2<MaxInputDependencies;Index2++)
					{
					if ( BranchInstructions[Index].InputDependencies[Index2] )
					if ( !theCompletionUnit->IsCompleted (
						BranchInstructions[Index].InstructionCount -
						BranchInstructions[Index].InputDependencies[Index2] ) )
						ReadyToIssue = false;
					}

				if ( ReadyToIssue )
					{
					if ( (FU == 0) && (BranchInstructions[Index].Memoizable) )
						{
						if(theCompletionUnit->CompleteInstruction (
							BranchInstructions [Index].InstructionCount ))
							{
							if (BranchInstructions [Index].IsAMissSpeculationDueToBranch)
								{
								(*ResetBranch) = true;
								if((BranchInstructions[Index].InstructionCount+1)<(*ToCount))
									{
									(*ToCount) = BranchInstructions [Index].InstructionCount + 1;
									(*ToAddress) = BranchInstructions [Index].FixupTarget;
									(*ResetDueToBranch) = true;
									}
								}
							memcpy ( & BranchInstructions [ Index ],
								 & BranchInstructions [ Index + 1],
								sizeof ( Instruction ) * (BranchQueueValid - Index - 1) );
							memcpy ( & BranchAddresses [ Index ],
								 & BranchAddresses [ Index + 1],
								sizeof ( uint ) * (BranchQueueValid - Index - 1) );
							--BranchQueueValid;
							}
						++Index;
						continue;
						}
					else
						{
						InjectPipeline(PipelineOfBranch, BranchInstructions, BranchAddresses, 1.0, BranchQueueValid);
						break;
						}
					}

				// add some value speculation code?

				++Index;
				}
			if(!Issued)
				{
				// o.k., go through the queue again, but this time searching for someone to value
				// speculate about... (that is value speculate about the inputs)!

				Index = 0;
				while(Index<BranchQueueValid)
					{
					ReadyToIssue = true;
					Confidence = 1.0;
					for(Index2=0;Index2<MaxInputDependencies;Index2++)
						{
						if( BranchInstructions[Index].InputDependencies[Index2] )
							{
							if ( !theCompletionUnit->IsCompletedWithConfidence (
								BranchInstructions[Index].InstructionCount -
								BranchInstructions[Index].InputDependencies[Index2],
								ValuePredictionConfidenceThreshold,
								&iConfidence) )
								ReadyToIssue = false;
							else
								Confidence = Confidence * iConfidence;
							}
						}
					if ( ReadyToIssue )
					if ( Confidence >= ValuePredictionConfidenceThreshold )
						{
						InjectPipeline(PipelineOfBranch,
							BranchInstructions, BranchAddresses, Confidence, BranchQueueValid);
						break;
						}
					++Index;
					}
				}
			}
		}

	return ( false );
	}

bool
ExecutionUnit	::	Flush (
	int64		ToCount )
	{
	// flush any pipeline, and queued instructons >= ToCount

	Stalled += StallPenaltyForBranchMissPredict;

	int	Index, FU;

	// flush queue
	Index = 0;
	while ( Index < IntegerQueueValid)
		{
		if ( IntegerInstructions [ Index ].InstructionCount >= ToCount )
			{
			memcpy ( & IntegerInstructions [ Index ],
					 & IntegerInstructions [ Index + 1],
					sizeof ( Instruction ) * (IntegerQueueValid - Index - 1) );
			memcpy ( & IntegerAddresses [ Index ],
					 & IntegerAddresses [ Index + 1],
					sizeof ( uint ) * (IntegerQueueValid - Index - 1) );
			--IntegerQueueValid;
			continue;
			}
		++Index;
		}
	// flush pipeline
	for(FU=0;FU<IntegerUnits;FU++)
		{
		Index = 0;
		for(Index=0;Index<IntegerPipelineDelay;Index++)
			{
			if ( PipelineOfInteger[FU][ Index ].Valid )
			if ( PipelineOfInteger[FU][ Index ].theInstruction.InstructionCount >= ToCount )
				PipelineOfInteger[FU][ Index].Valid = false;
			}
		}

	// flush queue
	Index = 0;
	while ( Index < FPQueueValid)
		{
		if ( FPInstructions [ Index ].InstructionCount >= ToCount )
			{
			memcpy ( & FPInstructions [ Index ],
					 & FPInstructions [ Index + 1],
					sizeof ( Instruction ) * (FPQueueValid - Index - 1) );
			memcpy ( & FPAddresses [ Index ],
					 & FPAddresses [ Index + 1],
					sizeof ( uint ) * (FPQueueValid - Index - 1) );
			--FPQueueValid;
			continue;
			}
		++Index;
		}
	// flush pipeline
	for(FU=0;FU<FPUnits;FU++)
		{
		Index = 0;
		for(Index=0;Index<FPPipelineDelay;Index++)
			{
			if ( PipelineOfFP[FU][ Index ].Valid )
			if ( PipelineOfFP[FU][ Index ].theInstruction.InstructionCount >= ToCount )
				PipelineOfFP[FU][ Index].Valid = false;
			}
		}

	// flush queue
	Index = 0;
	while ( Index < LoadQueueValid)
		{
		if ( LoadInstructions [ Index ].InstructionCount >= ToCount )
			{
			memcpy ( & LoadInstructions [ Index ],
					 & LoadInstructions [ Index + 1],
					sizeof ( Instruction ) * (LoadQueueValid - Index - 1) );
			memcpy ( & LoadAddresses [ Index ],
					 & LoadAddresses [ Index + 1],
					sizeof ( uint ) * (LoadQueueValid - Index - 1) );
			--LoadQueueValid;
			continue;
			}
		++Index;
		}
	// flush pipeline
	for(FU=0;FU<LoadUnits;FU++)
		{
		Index = 0;
		for(Index=0;Index<LoadPipelineDelay;Index++)
			{
			if ( PipelineOfLoad[FU][ Index ].Valid )
			if ( PipelineOfLoad[FU][ Index ].theInstruction.InstructionCount >= ToCount )
				PipelineOfLoad[FU][ Index ].Valid = false;
			}
		}

	// flush queue
	Index = 0;
	while ( Index < StoreQueueValid)
		{
		if ( StoreInstructions [ Index ].InstructionCount >= ToCount )
			{
			memcpy ( & StoreInstructions [ Index ],
					 & StoreInstructions [ Index + 1],
					sizeof ( Instruction ) * (StoreQueueValid - Index - 1) );
			memcpy ( & StoreAddresses [ Index ],
					 & StoreAddresses [ Index + 1],
					sizeof ( uint ) * (StoreQueueValid - Index - 1) );
			--StoreQueueValid;
			continue;
			}
		++Index;
		}
	// flush pipeline
	for(FU=0;FU<StoreUnits;FU++)
		{
		Index = 0;
		for(Index=0;Index<StorePipelineDelay;Index++)
			{
			if ( PipelineOfStore[FU][ Index ].Valid )
			if ( PipelineOfStore[FU][ Index ].theInstruction.InstructionCount >= ToCount )
				PipelineOfStore[FU][ Index ].Valid = false;
			}
		}
			
	// flush queue
	Index = 0;
	while ( Index < BranchQueueValid)
		{
		if ( BranchInstructions [ Index ].InstructionCount >= ToCount )
			{
			memcpy ( & BranchInstructions [ Index ],
					 & BranchInstructions [ Index + 1],
					sizeof ( Instruction ) * (BranchQueueValid - Index - 1) );
			memcpy ( & BranchAddresses [ Index ],
					 & BranchAddresses [ Index + 1],
					sizeof ( uint ) * (BranchQueueValid - Index - 1) );
			--BranchQueueValid;
			continue;
			}
		++Index;
		}
	// flush pipeline
	for(FU=0;FU<BranchUnits;FU++)
		{
		Index = 0;
		for(Index=0;Index<BranchPipelineDelay;Index++)
			{
			if ( PipelineOfBranch[FU][ Index ].Valid )
			if ( PipelineOfBranch[FU][ Index ].theInstruction.InstructionCount >= ToCount )
				PipelineOfBranch[FU][ Index ].Valid = false;
			}
		}
			
	return ( true );
	}

bool
ExecutionUnit	::	Dump ()
	{
	printf ( "QueueSizes : %d %d %d %d %d\n",
		IntegerQueueSize, FPQueueSize, LoadQueueSize, StoreQueueSize, BranchQueueSize );
	printf ( "QueueValid : %d %d %d %d %d\n",
		IntegerQueueValid, FPQueueValid, LoadQueueValid, StoreQueueValid, BranchQueueValid);

	int	Index = 0;
	printf ( "IQueue\n");
	for(Index=0;Index<IntegerQueueValid;Index++)
		{
		printf("%lld - %d, %d\n",
			IntegerInstructions[Index].InstructionCount,
			IntegerInstructions[Index].InputDependencies[0],
			IntegerInstructions[Index].InputDependencies[1]);
		}
	printf ( "FPQueue\n");
	for(Index=0;Index<FPQueueValid;Index++)
		{
		printf("%lld - %d, %d\n",
			FPInstructions[Index].InstructionCount,
			FPInstructions[Index].InputDependencies[0],
			FPInstructions[Index].InputDependencies[1]);
		}
	printf ( "LoadQueue\n");
	for(Index=0;Index<LoadQueueValid;Index++)
		{
		printf("%lld - %d, %d\n",
			LoadInstructions[Index].InstructionCount,
			LoadInstructions[Index].InputDependencies[0],
			LoadInstructions[Index].InputDependencies[1]);
		}

	printf ( "StoreQueue\n");
	for(Index=0;Index<StoreQueueValid;Index++)
		{
		printf("%lld - %d, %d\n",
			StoreInstructions[Index].InstructionCount,
			StoreInstructions[Index].InputDependencies[0],
			StoreInstructions[Index].InputDependencies[1]);
		}

	printf ( "BrancheQueue\n");
	for(Index=0;Index<BranchQueueValid;Index++)
		{
		printf("%lld - %d, %d\n",
			BranchInstructions[Index].InstructionCount,
			BranchInstructions[Index].InputDependencies[0],
			BranchInstructions[Index].InputDependencies[1]);
		}



	return ( true );
	}

