
class	CompletionUnit;

typedef	struct	_PipelineStage
	{
	bool		Valid;
	Instruction	theInstruction;
	uint		theAddress;
	double		Confidence;
	int64		Dependence0;
	int64		Dependence1;
	}	PipelineStage;

class	ExecutionUnit
	{
	public:
		// Statistics that are gathered...
	public:
		// Queues of incoming instructions
		Instruction	*IntegerInstructions;
		uint		*IntegerAddresses;
		Instruction	*FPInstructions;
		uint		*FPAddresses;
		Instruction	*LoadInstructions;
		uint		*LoadAddresses;
		Instruction	*StoreInstructions;
		uint		*StoreAddresses;
		Instruction	*BranchInstructions;
		uint		*BranchAddresses;

		// Pipeline stages
		PipelineStage	**PipelineOfInteger;
		PipelineStage	**PipelineOfFP;
		PipelineStage	**PipelineOfLoad;
		PipelineStage	**PipelineOfStore;
		PipelineStage	**PipelineOfBranch;
		
		int			IntegerQueueSize, FPQueueSize, LoadQueueSize,
					StoreQueueSize, BranchQueueSize;
		int			IntegerQueueValid, FPQueueValid, LoadQueueValid,
					StoreQueueValid, BranchQueueValid;
		int			IntegerUnits, FPUnits, LoadUnits, StoreUnits, BranchUnits;
		int			IntegerPipelineDelay, FPPipelineDelay, LoadPipelineDelay,
					StorePipelineDelay, BranchPipelineDelay;

		int			Stalled;
		int			L1MissPenalty, L2MissPenalty;

		CompletionUnit	*theCompletionUnit;

		bool		Initialize (
				int		aIntegerUnits,
				int		aFPUnits,
				int		aLoadUnits,
				int		aStoreUnits,
				int		aBranchUnits,

				int		aIntegerQueueSize,
				int		aFPQueueSize,
				int		aLoadQueueSize,
				int		aStoreQueueSize,
				int		aBranchQueueSize,

				int		aIntegerPipelineDelay,
				int		aFPPipelineDelay,
				int		aLoadPipelineDelay,
				int		aStorePipelineDelay,
				int		aBranchPipelineDelay,

				CompletionUnit	*aCompletionUnit,
				int				L1MissPenalty,
				int				L2MissPenalty	);

		int			StallsDueToCache (
				NormalDistribution	l1,
				NormalDistribution	l2 );

		bool		FreeSlots (
				int		*IntegerSlots,
				int		*FPSlots,
				int		*LoadSlots,
				int		*StoreSlots,
				int		*BranchSlots );

		bool		AddInstructions (
				Instruction		*aIntegerInstructions,
				uint			*aIntegerAddresses,
				int				NumberOfIntegerInstructions,
				Instruction		*aFPInstructions,
				uint			*aFPAddresses,
				int				NumberOfFPInstructions,
				Instruction		*aLoadInstructions,
				uint			*aLoadAddresses,
				int				NumberOfLoadInstructions,
				Instruction		*aStoreInstructions,
				uint			*aStoreAddresses,
				int				NumberOfStoreInstructions,
				Instruction		*aBranchInstructions,
				uint			*aBranchAddresses,
				int				NumberOfBranchInstructions );

		bool		SimulateOneClock (
				bool			*ResetBranch,
				uint			*ToAddress,
				int64			*ToCount,
				bool			*ResetDueToBranch );

		bool		Dump ();

		bool		Flush(
				int64			ToCount );
	};

