#	include	"Standard.h"

// 0, 0 means no trace cache, 1.0, 0 means perfect trace cache
NormalDistribution	TraceCacheBasicBlockContinueRate	=	{	0.0, 0.0	};

bool
FetchUnit	::	Initialize (
	BasicBlock	*CodeSpace,
	uint		aL1MissPenalty,
	uint		aL2MissPenalty)
	{
	theCode = CodeSpace;
	FetchProgramCounter = 0;
	CurrentBlock = CodeSpace;
	FetchingIncorrectly = false;
	Stalled = 0;
	L1MissPenalty = aL1MissPenalty;
	L2MissPenalty = aL2MissPenalty;
	Address = 0;
	CorrectBranchPredictions = IncorrectBranchPredictions = 0;
	EnsureHit = 0;
	FreeHits = 0;
	return ( true );
	}

bool
FetchUnit	::	EnsureCacheHit(
	int64	NextNInstructions)
	{
	if ( EnsureHit > NextNInstructions )
		return ( true );
	EnsureHit = NextNInstructions;
	return ( true );
	}

bool
FetchUnit	::	ResetFetchProgramCounter(
	uint	aAddress)
	{
	FetchProgramCounter = Address;
	FetchingIncorrectly = false;
	Stalled = 0;
	BasicBlock	*b = CurrentBlock;
	CurrentBlock = FindBasicBlock(aAddress);
	Address = aAddress;
	if(!CurrentBlock)
		{
		printf("CB = NULL!\n");
		printf("b->Address = %d\n", Address);
		printf("b->NumberOfInstructions = %d\n", b->NumberOfInstructions);
		printf("b->Next = %x\n", b->Next);
		CurrentBlock = theCode;
		Address = 0;
		}
	return ( true );
	}

BasicBlock	*
FetchUnit	::	FindBasicBlock (
	uint	aAddress)
	{
	BasicBlock	*b = theCode;

	int	Index = 0;
	while(b)
		{
		if(	(b->Address <= aAddress)&&
			((b->Address + b->NumberOfInstructions + 1) > aAddress))
			return(b);
		b = b->Next;
		++Index;
		}
	return(NULL);
	}

int
FetchUnit	::	StallsDueToCache(
	NormalDistribution	l1,
	NormalDistribution	l2)
	{
	// generate a uniform random number and a normally distributed
	// number.  if uniform <= normally distributed then a hit.
	// otherwise a miss!

	if ( EnsureHit > 0 )
		{
		--EnsureHit;
		++FreeHits;
		return ( 0 );
		}
	if (	RandomVariable :: GetUniformRandomVariable(1.0) <=
			RandomVariable :: GetNormalRandomVariable(l1)	)
		{
	//	if ( FreeHits )
	//		{
	//		--FreeHits;
	//		if (	RandomVariable :: GetUniformRandomVariable(1.0) <=
	//				RandomVariable :: GetNormalRandomVariable(l1)	)
	//			return(0);
	//		}
	//	else
			return (0);
		}

	// Miss in L1!
	if (	RandomVariable :: GetUniformRandomVariable(1.0) <=
			RandomVariable :: GetNormalRandomVariable(l2)	)
		return ( L1MissPenalty + 1  );

	return ( L2MissPenalty + L1MissPenalty + 1);
	}


bool
FetchUnit	::	SimulateOneClock (
	Instruction	*Instructions,
	uint		*Addresses,
	int			*FetchedInstructions,
	int			MaxFetchSize)
	{
	bool	Predicted;
	bool	GoNext;

	if ( Stalled )	// from cache miss
		{
		--Stalled;
		(*FetchedInstructions) = 0;
		return(true);
		}
	// O.k.; determine whether there is going to be a cache miss
	// on any of these instructions...
	int	Index = 0;
	while ( Index < MaxFetchSize )
		{
		// offset
		int	Offset = Address - CurrentBlock->Address;

		//if(!Offset)
		//	{
			Stalled += StallsDueToCache (
				CurrentBlock->Instructions[0].L1InstructionHitPredictability,
				CurrentBlock->Instructions[0].L2InstructionHitPredictability);
			if(Stalled)
				{
				--Stalled;
				(*FetchedInstructions) = Index;
				return(true);
				}
		//	}
		if(Offset == CurrentBlock->NumberOfInstructions)
			{
			// wowa, this is the branch.
			// decide whether we are going to correctly predict this branch.
			// yes, we do this right away, which is some what backwards from how it
			// is computed.  But this is a HLS, so take advantage of it!
			//if (	RandomVariable :: GetUniformRandomVariable(2.7) <
			//		CurrentBlock->BranchPredictability )
			if (	RandomVariable :: GetUniformRandomVariable(1.0) >
					RandomVariable :: GetNormalRandomVariable(CurrentBlock->BranchPredictability) )
				{
				Predicted = true;
				++IncorrectBranchPredictions;
				}
			else
				{
				Predicted = false;
				++CorrectBranchPredictions;
				}

			(*Addresses) = Address;
			(*Instructions) = CurrentBlock->Instructions[0];
			(*Instructions).FunctionalUnitRequirements = Branch;
			(*Instructions).IsJump = true;
			(*Instructions).IsAMissSpeculationDueToBranch = FetchingIncorrectly;
			(*Instructions).IsAMissSpeculationDueToData = false;
			(*Instructions).ExecutionState = 0;
			(*Instructions).Memoizability = CurrentBlock->MemoizabilityOfBranch;
			if(FetchingIncorrectly == false)
				FetchingIncorrectly = Predicted;
			if(	RandomVariable::GetUniformRandomVariable(1.0) <= 0.5)
				{
				GoNext = true;
				CurrentBlock = CurrentBlock->BranchTargets[0];
				(*Instructions).FixupTarget = 
					CurrentBlock->BranchTargets[1]->Address;
				}
			else
				{
				GoNext = false;
				CurrentBlock = CurrentBlock->BranchTargets[1];
				(*Instructions).FixupTarget = 
					CurrentBlock->BranchTargets[0]->Address;
				}
			(*Instructions).InputDependencies[0] = CurrentBlock->BranchInputDependencies[0];
			(*Instructions).InputDependencies[1] = CurrentBlock->BranchInputDependencies[1];
			++Instructions;
			++Addresses;
			Address = CurrentBlock->Address;
			++Index;
			if ( RandomVariable :: GetNormalRandomVariable ( TraceCacheBasicBlockContinueRate ) >=
				 RandomVariable :: GetUniformRandomVariable ( 1.0 )	)
				continue;
			else
				{
				if ( ! GoNext )
					break;
				else
					continue;
				}
			}
		else
			{
			if(!Offset)
				{
				Stalled = StallsDueToCache (
					CurrentBlock->Instructions[Offset].L1InstructionHitPredictability,
					CurrentBlock->Instructions[Offset].L2InstructionHitPredictability);
				}
			(*Addresses) = Address;
			(*Instructions) = CurrentBlock->Instructions[Offset];
			(*Instructions).IsJump = false;
			(*Instructions).IsAMissSpeculationDueToBranch = FetchingIncorrectly;
			(*Instructions).IsAMissSpeculationDueToData = false;
			(*Instructions).ExecutionState = 0;
			// merge load/store units
			//if ((*Instructions).FunctionalUnitRequirements == Store)
			//	(*Instructions).FunctionalUnitRequirements = Load;
				
			++Instructions;
			++Addresses;
			}
		++Index;
		++Address;
		}
	(*FetchedInstructions) = Index;
	return(true);
	}

