#	include	"Standard.h"


bool
CompletionUnit	::	Initialize (
	int		aMaxCompletionsPerCycle)
	{
	Completed = -1;
	MaxCompletionsPerCycle = aMaxCompletionsPerCycle;
	PredictionCount = 0;
	return ( true );
	}

bool
CompletionUnit	::	SimulateOneClock ()
	{
	CompletedThisCycle = 0;
	}

bool
CompletionUnit	::	CompletePrediction (
	int64	InstructionCount,
	double	Confidence)
	{
	// check to see if it has already been completed..
	if ( IsCompleted(InstructionCount, true) )
		{
		printf("Completion Unit : CompletePrediction -- but already completed! %lld\n", InstructionCount);
		return ( false );	// already completed?  might print something here...
		}

	// add the prediction
	SpeculativelyCompletedInstruction	SCI;

	SCI.theInstructionNumber = InstructionCount;
	SCI.Confidence = Confidence;
	SCI.DependentUpon0 = -1;
	SCI.DependentUpon1 = -1;

	Predictions.Add(SCI);
	++PredictionCount;
	return ( true );
	}

bool
CompletionUnit	::	CompleteInstruction (
	int64	theInstruction,
	double	Confidence,
	int64	DependentUpon0,
	int64	DependentUpon1,
	bool	MustComplete)
	{
	//printf("Completing : %lld - Completed = %lld\n", theInstruction, Completed );
	int	Index = 0;

	if ( !MustComplete )
	if ( MaxCompletionsPerCycle > 0 )
	if ( CompletedThisCycle >= MaxCompletionsPerCycle )
		return ( false );


	Index = 0;	// check to see if we actually are completing a follow up to a prediction
	while ( Index < Predictions.GetSize() )
		{
		if ( Predictions[Index].theInstructionNumber == theInstruction )
			{
			SpeculativelyCompletedInstruction	SCI = Predictions[Index];
			Predictions.RemoveAt(Index);
			CompleteInstruction(theInstruction, 1.0, -1, -1, true);
			// now empty followup array of anything that can be emptied
			while ( CompleteFollowUpInstructions () )
				;
			return ( true );
			}
		++Index;
		}

	if ( theInstruction < Completed )
		{
		printf("CompletionUnit : Error %s/%d - %lld\n", __FILE__, __LINE__, theInstruction);
		//return ( false );
		}
	++CompletedThisCycle;
	// check to see if this instruction is being completed by a prediction
	// if so, then we add it to the follow up instruction.
	if ( (!IsCompleted(DependentUpon0))||(!IsCompleted(DependentUpon1)) )
		{
		SpeculativelyCompletedInstruction		SCI;

		SCI.theInstructionNumber = theInstruction;
		SCI.Confidence = Confidence;
		SCI.DependentUpon0 = DependentUpon0;
		SCI.DependentUpon1 = DependentUpon1;
		FollowUpInstructions.Add(SCI);
		/*printf("Completing Follow on for %lld: %lld, %lld, %d, %d\n",
			theInstruction,
			DependentUpon0, DependentUpon1,
			IsCompleted(DependentUpon0),
			IsCompleted(DependentUpon1));
		*/
		return ( true );
		}
	if ( theInstruction == (Completed + 1) )
		{
		++Completed;
		// check to see if we can migrate anything from the array
		while ( CompletedInstructions.GetSize () )
			{
			if ( CompletedInstructions[0] == (Completed + 1) )
				{
				++Completed;
				CompletedInstructions.RemoveAt(0);
				continue;
				}
			break;
			}
		return ( true );
		}
	// dump completed instruction into array
	Index = 0;
	while ( Index < CompletedInstructions.GetSize () )
		{
		if ( theInstruction < CompletedInstructions[Index] )
			{
			CompletedInstructions.InsertAt(theInstruction, Index);
			return(true);
			}
		++Index;
		}
	CompletedInstructions.Add(theInstruction);
	return ( true );
	}

bool
CompletionUnit	::	Flush (
	int64	ToInstruction )
	{
	if ( ToInstruction <= Completed )
		Completed = ToInstruction-1;

	int	Index = CompletedInstructions.GetSize() - 1;
	while ( Index >= 0 )
		{
		if ( ToInstruction <= CompletedInstructions[Index] )
			CompletedInstructions.RemoveAt(Index);
		else
			break;
		--Index;
		}
	// flush any predictions <= ToInstruction
	Index = 0;
	while ( Index < Predictions.GetSize() )
		{
		if ( ToInstruction <= Predictions[Index].theInstructionNumber )
			Predictions.RemoveAt(Index);
		else
			++Index;
		}
	Index = 0;
	while ( Index < FollowUpInstructions.GetSize() )
		{
		if ( ToInstruction <= FollowUpInstructions[Index].theInstructionNumber )
			FollowUpInstructions.RemoveAt(Index);
		else
			++Index;
		}
	return ( true );
	}

bool
CompletionUnit	::	IsCompletedWithNoSpeculation (
	int64	theInstruction )
	{
	if ( theInstruction <= Completed )
		return ( true );

	// the code runs faster if we search down instead of up.  Nature of the code gen though...
/*
	int	Index = 0;
	while ( Index < CompletedInstructions.GetSize() )
		{
		if ( theInstruction == CompletedInstructions [ Index ] )
			return ( true );
		if ( theInstruction < CompletedInstructions [ Index ] )
			return ( false );
		++Index;
		}
*/
	if(!CompletedInstructions.GetSize())
		return ( false );
	int	Index = CompletedInstructions.GetSize() - 1;
	while( Index >= 0 )
		{
		if ( theInstruction == CompletedInstructions [ Index ] )
			return ( true );
		if ( theInstruction > CompletedInstructions [ Index ] )
			return ( false );
		--Index;
		}
	return ( false );
	}

bool
CompletionUnit	::	IsCompleted (
	int64	theInstruction,
	bool	UsePredicted,
	bool	*Speculatively,
	double	*Confidence )
	{
	if ( Speculatively )	(*Speculatively) = false;
	if ( Confidence )		(*Confidence) = 1.0;
	
	if ( IsCompletedWithNoSpeculation ( theInstruction ) )
		return ( true );

	if ( ! UsePredicted )
		return ( false );

	// Search all completed predictions
	// and all follow up predictions
	int	Index = 0;
	while ( Index < Predictions.GetSize() )
		{
		if ( Predictions[Index].theInstructionNumber == theInstruction )
			{
			if( Speculatively )		(*Speculatively) = true;
			if( Confidence )		(*Confidence) = Predictions[Index].Confidence;
			return ( true );
			}
		++Index;
		}
	Index = 0;
	while ( Index < FollowUpInstructions.GetSize() )
		{
		if ( FollowUpInstructions[Index].theInstructionNumber == theInstruction )
			{
			if ( Speculatively )	(*Speculatively) = true;
			if ( Confidence )		(*Confidence) = FollowUpInstructions[Index].Confidence;
			return ( true );
			}
		++Index;
		}
	return ( false );
	}

bool
CompletionUnit	::	CompleteFollowUpInstructions ()
	{
	bool	Change = false;

	int		Index = 0 ;
	while ( Index < FollowUpInstructions.GetSize() )
		{
		if ( IsCompleted(FollowUpInstructions[Index].DependentUpon0, false, NULL, NULL))
		if ( IsCompleted(FollowUpInstructions[Index].DependentUpon1, false, NULL, NULL))
			{
			// complete instruction!
			SpeculativelyCompletedInstruction	SCI = FollowUpInstructions[Index];
			FollowUpInstructions.RemoveAt(Index);
			CompleteInstruction(SCI.theInstructionNumber, 1, -1, -1, true);
			Change = true;
			continue;
			}
		++Index;
		}
	return ( Change );
	}

		
bool
CompletionUnit	::	IsCompletedWithConfidence(
	int64	theInstruction,
	double	WithConfidence,
	double	*theConfidence)
	{
	double	Confidence;
	bool	Speculatively;
	if ( !IsCompleted(theInstruction, true, &Speculatively, &Confidence) )
		return ( false );

	if ( Confidence < WithConfidence )
		return ( false );

	if ( theConfidence )	(*theConfidence) = Confidence;

	return ( true );	
	}

