#	include	"Standard.h"

double	ValuePredictionConfidenceThreshold	=	0.9;

bool
DispatchUnit	::	Initialize (
	int		aWindowSize,
	int		aMaxDispatchAmount,
	CompletionUnit	*aCompletionUnit )
	{
	WindowSize = aWindowSize;
	Instructions = new Instruction[WindowSize];
	Addresses = new uint[WindowSize];
	Valid = 0;
	InstructionCount = 0;
	MaxDispatchAmount = aMaxDispatchAmount;
	theCompletionUnit = aCompletionUnit;
	return ( true );
	}

bool
DispatchUnit	::	AddInstructions (
	Instruction	*aInstructions,
	uint		*aAddresses,
	uint		Size )
	{
	if ( Size > FreeSlots () )	return ( false );

	memcpy ( & Instructions [ Valid ], aInstructions,
		sizeof ( Instruction ) * Size );
	memcpy ( & Addresses [ Valid ], aAddresses,
		sizeof ( uint ) * Size );

	// now bind dependencies dynamically.
	int	Index = Valid;
	while ( Index < (Valid + Size) )
		{
		// order the instruction
		Instructions [ Index ].InstructionCount = InstructionCount++;
		// determine how well we do on predicting it... (value)
		if ( RandomVariable :: GetUniformRandomVariable(1.0) <=
			 RandomVariable :: GetNormalRandomVariable(
				Instructions [ Index ].OutcomePredictability) )
			Instructions [ Index ].MissPredictedOutcome = false;
		else
			{Instructions [ Index ].MissPredictedOutcome = true;
			//printf("Miss Predicted outcome: %f,%f\n",
			//	Instructions[Index].OutcomePredictability.u,
			//	Instructions[Index].OutcomePredictability.o);}
			}
		// determine how confident we are in this prediction...
		Instructions[Index].OutcomePredictionConfidence =
			RandomVariable :: GetNormalRandomVariable (
				Instructions [ Index ].OutcomeConfidence );
		// determine if we are going to be able to memoize it.
		if ( RandomVariable :: GetUniformRandomVariable(1.0) <=
			 RandomVariable :: GetNormalRandomVariable(
				Instructions [ Index ].Memoizability) )
			Instructions [ Index ].Memoizable = true;
		else
			Instructions [ Index ].Memoizable = false;
		++Index;
		}
	Valid += Size;
	return ( true );
	}

bool
DispatchUnit	::	Flush (
	int64	aInstructionCount)
	{
	int	Index = 0;
	while (Index < Valid )
		{
		if ( Instructions [ Index ].InstructionCount >= aInstructionCount )
			break;
		++Index;
		}
	Valid = Index;
	InstructionCount = aInstructionCount;
	return ( true );
	}

int
DispatchUnit	::	FreeSlots ()
	{
	return ( WindowSize - Valid );
	}

// Note: If we have > 50% knowledge then we can get to peek at the outcome!  If we have < 50%
// knowledge, then well we have to roll our dice with whatever the predictions are.  Also note that
// none of this happens if we are not very confident in our predictor, so there is a first order effect
// of the confidence threshhold
#	define	OutputInstruction(xxx,yyy,iii)	\
			if(Instructions[iii].OutcomePredictionConfidence >= ValuePredictionConfidenceThreshold)\
				{\
				if(RandomVariable::GetNormalRandomVariable(Instructions[iii].OutcomeKnowledge) >= 0.50)\
					{\
					if ( Instructions[iii].MissPredictedOutcome == false)\
						theCompletionUnit->CompletePrediction(\
							Instructions[iii].InstructionCount,\
							Instructions[iii].OutcomePredictionConfidence);\
					else\
						Instructions[iii].MissPredictedOutcome = false;\
					}\
				else\
					{\
					theCompletionUnit->CompletePrediction(\
						Instructions[iii].InstructionCount,\
						Instructions[iii].OutcomePredictionConfidence);\
					}\
				}\
			else\
				Instructions[iii].MissPredictedOutcome = false;\
			memcpy(xxx,						\
				&Instructions [ iii ],		\
				sizeof ( Instruction ) );	\
			memcpy(yyy,						\
				&Addresses [ iii ],			\
				sizeof ( uint ) );			\
			memcpy(&Instructions [ iii ],	\
					&Instructions [ iii + 1],	\
					(Valid - iii - 1)*sizeof(Instruction));		\
			memcpy(&Addresses [ iii ],	\
					&Addresses [ iii + 1],	\
					(Valid - iii - 1)*sizeof(uint));		\
			++xxx;++yyy;--Valid;++Dispatched;


bool
DispatchUnit	::	SimulateOneClock (
	Instruction	*aIntegerInstructions,
	uint		*aIntegerAddresses,
	Instruction	*aFPInstructions,
	uint		*aFPAddresses,
	Instruction	*aLoadInstructions,
	uint		*aLoadAddresses,
	Instruction	*aStoreInstructions,
	uint		*aStoreAddresses,
	Instruction	*aBranchInstructions,
	uint		*aBranchAddresses,
	int			*MDSInteger,
	int			*MDSFP,
	int			*MDSLoad,
	int			*MDSStore,
	int			*MDSBranch)
	{
	// dispatch up to n instructions of each...
	// try and get
	int	Index;
	int	Dispatched = 0;
	Index = 0;
	while ( (Index < Valid) )
		{
		if ( Dispatched >= MaxDispatchAmount )
			return ( true );
		// first deterime if we can predict it!
		if ( (*MDSInteger) > 0 )
		if ( Instructions [ Index ].FunctionalUnitRequirements == Integer )
			{
			OutputInstruction(aIntegerInstructions, aIntegerAddresses, Index);
			--(*MDSInteger);
			continue;
			}
		if ( (*MDSFP) > 0 )
		if ( Instructions [ Index ].FunctionalUnitRequirements == FloatingPoint )
			{
			OutputInstruction(aFPInstructions, aFPAddresses, Index);
			--(*MDSFP);
			continue;
			}
		if ( (*MDSLoad) > 0 )
		if ( Instructions [ Index ].FunctionalUnitRequirements == Load )
			{
			OutputInstruction(aLoadInstructions, aLoadAddresses, Index);
			--(*MDSLoad);
			continue;
			}
		if ( (*MDSStore) > 0 )
		if ( Instructions [ Index ].FunctionalUnitRequirements == Store )
			{
			OutputInstruction(aStoreInstructions, aStoreAddresses, Index);
			--(*MDSStore);
			continue;
			}
		if ( (*MDSBranch) > 0 )
		if ( Instructions [ Index ].FunctionalUnitRequirements == Branch )
			{
			OutputInstruction(aBranchInstructions, aBranchAddresses, Index);
			--(*MDSBranch);
			continue;
			}
		++Index;
		}
	return ( true );
	}

bool
DispatchUnit	::	Dump ()
	{
	printf("Dispatch Unit - Valid = %d of %d\n",
		Valid, WindowSize );
	int	Index;
	for(Index=0;Index<Valid;Index++)
		{
		printf("%d: %lld (%d) - %d,%d\n", Index, Instructions[Index].InstructionCount,
			Instructions[Index].FunctionalUnitRequirements,
			Instructions[Index].InputDependencies[0],
			Instructions[Index].InputDependencies[1]);
		}
	return(true);		
	}

